/*
#### Copyright (c) 1998 Pacific HiTech, Inc.
####              All Rights Reserved
####              Distributable under the terms of the GPL.  
#### File:        autowin.c 
#### Author:      koji Ashida (koji@pht.com) #### Version:     1.2J
#### Date:        Febrary 16, 1998
#### Description:
####    Automatic treatment of windows.  This should help the localization
####    work by not needing the explicit coordinates and such.
####
#### ChangeLog:
####    (none)
*/

#include "autowin.h"
#include <errno.h>

//////////////////////////////////////////////////////////////////////
// data structures and declarations.
//////////////////////////////////////////////////////////////////////

/*   define the extra screen area used.   */
#define MARGINE_FACTOR .90

// used for japanese handling.
#define JPN_PUNCT_NOSTART \
   "                          " \
   "     "
#define JPN_PUNCT_NOEND   "        "

/*   data structure taken from newt_pr.h of newt source package.   */
struct newtComponent_struct {
    /* common data */
    int height, width; 
    int top, left;
    int takesFocus;

    struct componentOps * ops;

    newtCallback callback;
    void * callbackData;

    void * data;
} ;


/*   data structure to group newt components together.   */
typedef struct componentGroup_struct {
    int height, width;
    enum cg_orientation orientation;
    enum cg_gravity gravity;
    int numComponents;
    newtComponent *components;
} componentGroup_struct_type;




//////////////////////////////////////////////////////////////////////
// text manipulation functions.
//////////////////////////////////////////////////////////////////////

/*   find out the number of rows and columns necessary to display
     this text.   note that if the text is zero length, it still
     requires 1x1 area.  similarly, if the end of the text is
     newline, extra row will be reserved.   */
void textDimensions(char *text, int *x, int *y)
{
    int rows      = 1;
    int cols      = 1;
    int this_col  = 0;

    while(*text)
    {
        this_col++;

        if(*text == '\n')
        {
            /*   end of line happens with no regard to the language.   */
            if(this_col > cols)
            {
                cols = this_col;
            }
            this_col = 0;
            rows++;
        }
        text++;
    }

    if(this_col > cols)
    {
      cols = this_col;
    }

    *x = cols;
    *y = rows;
}


// this function catenates a line into a buffer.
// return the number of characters processed.
int addALine(char *dest, char *src, int width, bool *endable)
{
    int i;

    // check if there is a newline.
    for(i = 0; src[i] && i < width; i++)
    {
        if(src[i] == '\n')
        {
            // split the line here.
            strncat(dest, src, i + 1);
            return i + 1;
        }
    }

    // check if we have a whole line full of length.
    if(i < width)
    {
	    /*
        // seems to be the last line of a paragraph.  we don't have
        // to split.
	*/
        strcat(dest, src);
        return i + 1;
    }

    // check where to break from backwards now.
    for(i = width - 1; i >= 0; i--)
    {
        if(endable[i])
        {
            // we can split the line here.
            strncat(dest, src, i + 1);
            strcat(dest, "\n");
            return i + 1;
        }
    }

    /*
    // we have no choice but to break the line wherever.
    // we can at most proceed widht - 1 since width includes
    // a '\n'.
    */
    strncat(dest, src, width);
    strcat(dest, "\n");
    return width;
}

/*
// break a long line into smaller pieces.
// this takes a text buffer and max length of a line.
// it'll set y value and return a new buffer which has newline
// embedded.  the passed buffer is deleted automatically.
*/
char *breakLines(char *text, int x, int *y)
{
    int text_len, i;
    char *buf;
    bool *endable;

    // prepare a workspace.
    text_len = strlen(text);
    endable = malloc(sizeof(bool) * text_len);
    buf = malloc(sizeof(char) * text_len * 2);
    buf[0] = '\0';
    // we first check which characters of the string can be the last
    // character of a line.
    i = 0; 
    while(i < text_len)
    {
	    /*
        // if it's a first byte of japanese char, it can't be the
        // last byte of a line.
	*/
        if((unsigned char)text[i] >= 0xa0 && (unsigned char)text[i] <= 0xff)
        {
            // it was a first byte of japanese char.
            endable[i++] = FALSE;

            // check the second char.
            if(i < text_len)
            {
                char kanji[3];
                kanji[0] = text[i - 1];
                kanji[1] = text[i];
                kanji[2] = '\0';
                if(strstr(JPN_PUNCT_NOSTART, kanji))
                {
			/*
                    // this is a japanese punctuation character
                    // which shouldn't start a line.
		    */
                    if(i >= 2)
                    {
                        endable[i - 2] = FALSE;
                    }
                    endable[i++] = TRUE;
                }
                else if(strstr(JPN_PUNCT_NOEND, kanji))
                {
			/*
                    // this is a japanese punctuation character which
                    // shouldn't end a line.
		    */
                    endable[i++] = FALSE;
                }
                else
                {
			/*
                    // otherwise, japanese string can be devided anywhere.
		    */
                    endable[i++] = TRUE;
                }
            }
        }
        else
        {
		/*
            // english char.  only allowed one is a space.
	    */
            endable[i++] = strchr(" \t\n", text[i]) ? TRUE : FALSE;
        }
	/*
        // i should have been incremented (at least one).
	*/
    }

    /*
    // devide the lines.
    */
    for(*y = 0, i = 0; i < text_len; (*y)++)
    {
        i += addALine(buf, text + i, x, endable + i);
    }

    free(text);
    return buf;
}


/*   make a good guess of text dimensions.  it is useful when
     the caller of msgWindow did not care to embet newlines
     in the text.   */
char *smartTextDimensions(char *text, int *x, int *y)
{
    int sc_x, sc_y;

    /*   we first try the regular method.   */
    textDimensions(text, x, y);

    /*   check if the result was acceptable.   */
    newtGetScreenSize(&sc_x, &sc_y);
    if((double)*x < (double)sc_x * MARGINE_FACTOR)
    {
        /*   ok.  we'll let it pass.   */
        return text;
    }

    /*   we'll have to insert newlines to break the long lines.
         this is because we can't trust newt to handle japanese
         character handling. 
         TODO: improve this rather rough estimate.
               it should be albe to take differenct widths 
               depending on the context.   */
    *x = (double)sc_x * MARGINE_FACTOR;
    return breakLines(text, *x, y);
}

/*
// make a text box and automatically determine the sizes.   */
newtComponent properTextbox(char * message, ...) 
{
    newtComponent text_box;
    char *buf = NULL;
    int size = 10000;
    int width, height;
    va_list args;

    // create the text first.
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    // find the dimensions and create the text box.
    buf = smartTextDimensions(buf, &width, &height);
    text_box = newtTextbox(0, 0, width + 1, height, NEWT_TEXTBOX_WRAP);
    newtTextboxSetText(text_box, buf);
    free(buf);

    return text_box;
}

/*
// this one draws text on the screen just like newtDrawRootText does.
// this one is convenient because it takes printf kinda arguments.
// negative left argument is treated differently from newtDrawText's.
// it'll adjust the right side margine of the text against the screen.
*/
int putRootText(int left, int top, char *message, ...)
{
    char *buf = NULL;
    int len;
    int size = 10000;
    va_list args;

    // create the text first.
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    // draw the text.
    len = strlen(buf);
    newtDrawRootText(top, left >= 0 ? left : -len, buf);

    return len;
}


// this one erase the text drawn before.
void eraseRootText(int left, int top, int len)
{
    char *space;

    // prepare a white space of specified length.
    space = alloca(len + 1);
    memset(space, ' ', len);
    space[len] = '\0';

    // draw it.
    newtDrawRootText(top, left >= 0 ? left : -len, space);
}
    
/*
//////////////////////////////////////////////////////////////////////
// list handling. 
//////////////////////////////////////////////////////////////////////
*/
/*
// add list items to the list.
*/
selectionList addToSelectionList(selectionList *list, ...)
{
    selectionList head = NULL;
    selectionList last = NULL;
    selectionList *head_ptr = &head;
    char *title = NULL;
    va_list args;
    bool is_title = TRUE;
    void *ptr;

    // go through the argument list.
    va_start(args, list);
    while((ptr = va_arg(args, void *)))
    {
        if(is_title)
        {
            // first item of the pair.
            title = (char *)ptr;
        }
        else
        {
            // second item of the pair.
            last = malloc(sizeof(selectionList_struct_type));
            last->title = title;
            last->data = ptr;
            *head_ptr = last;
            head_ptr = &(last->next);
        }

        // move on to the next one.
        is_title = !is_title;
    }
    va_end(args);

    // check if we had anything at all.
    if(!head || !last)
    {
        return list ? *list : NULL;
    }

    /*
    // make what we have a proper list.
    */
    head->last = last;
    head->last->next = NULL;

    /*
    // check if we have a list to append this.
    */
    if(list == NULL || *list == NULL)
    {
	    /*
        // we can return the one we've just made.
	*/
        *list = head;
        return head;
    }

    /*
    // we'll append our list to the existing one.
    */
    (*list)->last->next = head;
    (*list)->last = last;

    return *list;
}

/*
// make a compgroup looking struct of listbox.
*/
componentGroup selectionListbox(newtComponent *listboxComp,
    bool multi, int height, int def, selectionList list)
{
    selectListbox_struct_type *listbox;
    int listwidth = 1;
    int len;
    selectionList ptr;

    listbox = malloc(sizeof(selectListbox_struct_type));

    /*
    // find the longest string to figure out the listbox width.
    */
    ptr = list;
    while(ptr)
    {
        if((len = strlen(ptr->title)) > listwidth)
        {
            listwidth = len;
        }
        ptr = ptr->next;
    }

    /*
    // give two columns for the scroll bar.  also, it has to 
    // be marked with negation so we know it's a listbox later.
    */
    listbox->width = -(listwidth + 2);

    /*
    // set other fields.
    */
    listbox->height = height;
    listbox->def = def;
    listbox->list = list;
    listbox->comp = listboxComp;
    listbox->multi = multi;

    /*
    // pretending it's a component group, return.
    */
    return (componentGroup)listbox;
}
    

/*
//////////////////////////////////////////////////////////////////////
// generic window generation functions.
//////////////////////////////////////////////////////////////////////
*/

/*   open a centered window within the screen dimensions.   */
void openCenteredWindow(int width, int height, char *title)
{
    int sc_x, sc_y;

    newtGetScreenSize(&sc_x, &sc_y);
    newtOpenWindow((sc_x - width) / 2, (sc_y - height) / 2, 
        width, height, title);
}


/*
//   clean up the form which was opened with ask_keep option of askUser.
*/
void releaseForm(newtComponent *form)
{
    if(*form != NULL)
    {
        newtFormDestroy(*form);
        newtPopWindow();
    }
    *form = NULL;
}
    
/*
//  return the list of components grouped together.
*/
componentGroup compGroup(
    enum cg_orientation orientation, enum cg_gravity gravity, ...)
{
    componentGroup group;
    newtComponent comp;
    newtComponent *ptr;
    int numComps = 0;
    va_list args;
    
    /*   count the number of components.   */
    va_start(args, gravity);
    while((comp = va_arg(args, newtComponent)))
    { 
        numComps++; 
    }
    va_end(args);

    /*   prepare containers for those components.   */
    group = malloc(sizeof(componentGroup_struct_type));
    group->components = malloc(sizeof(newtComponent) * numComps);

    /*   set the data into this structure.   */
    group->width = group->height = 0;
    group->orientation = orientation;
    group->gravity = gravity;
    group->numComponents = numComps;
    ptr = group->components;
    va_start(args, gravity);
    while((comp = va_arg(args, newtComponent))) 
    { 
        /*   keep the object in the array.   */
        *(ptr++) = comp; 

        /*   compute widht and height.   */
        if(orientation == vertical)
        {
            group->height += abs(comp->height);
            if(group->width < abs(comp->width))
            {
                group->width = abs(comp->width);
            }
        }
        else /*   should be orientation == GRP_HORIZ.   */
        {
            if(group->height < abs(comp->height))
            {
                group->height = abs(comp->height);
            }
            group->width += abs(comp->width);
        }
    }
    va_end(args);

    /*   mark the object as a group by setting the negative number
         for width and height, and return this object.   */
    group->height *= -1;
    group->width *= -1;
    return group;
}

/*
// adjust the margine space.
*/
componentGroup groupMargine(int x_margine, int y_margine, componentGroup group)
{
	/*
    // check if it really is a group, or a component.
    */
    if(group->width < 0)
    {
	    /*
        // it was a group.
	*/
        group->width -= x_margine;
        group->height -= y_margine;
    }
    else
    {
	    /*
        // it actually was a component.
	*/
        group->width += x_margine;
        group->height += y_margine;
    }

    return group;
}


/*   add all the components in the group tree into this form.
     while doing that, we have to figure out the coordinates
     of the components.  also we have to delette the group 
     objects along the way and we have to do all these all
     in one shot.   */
void formAddGroup(newtComponent form, componentGroup group, int x, int y)
{
    /*   check what it actually is.   */
    if(group->width < 0 && group->height < 0)
    {
        /*   it was a group.  we have to go through each of them.   */
        int i;
        for(i = 0; i < group->numComponents; i++)
        {
            /*   figure out the coordinates.   */
            int left, top, comp_height, comp_width;
            newtComponent comp;

            comp = group->components[i];
            if(group->orientation == vertical)
            {
                // align along the x coordinate.
                // y coordinate is defined already.
                left = x;
                top  = y;
                if(group->gravity == center)
                {
                    left += (abs(group->width) - abs(comp->width)) / 2;
                }
                else if(group->gravity == bottomRight)
                {
                    left += abs(group->width) - abs(comp->width);
                }
            }
            else /*   when group->orientation == horizontal.   */
            {
                // x coordinate is defined already.
                // align along the y coordinate.
                left = x;
                top  = y;
                if(group->gravity == center)
                {
                    top += (abs(group->height) - abs(comp->height)) / 2;
                }
                else if(group->gravity == bottomRight)
                {
                    top += abs(group->height) - abs(comp->height);
                }
            }

            /*   with the information we just figured out, we can
                 move on to the recursive call.   */
            comp_height = comp->height;
            comp_width  = comp->width;
            formAddGroup(form, (componentGroup)comp, left, top);

            /*   check the coordinates of the next component.   */
            if(group->orientation == vertical)
            {
                y += abs(comp_height);
            }
            else /*   when group->orientation == horizontal.   */
            {
                x += abs(comp_width);
            }
        }

        /*   the use of this group as a container has just finished.
             we'll get rid of it.   */
        free(group->components);
        free(group);
    }
    else if(group->width < 0)
    {
        // it's supposed to be a listbox.
        newtComponent listbox;
        selectListbox_struct_type *lb = (selectListbox_struct_type *)group;
        listbox = newtListbox(x, y, lb->height, NEWT_LISTBOX_RETURNEXIT);

        // add all the items.
        while(lb->list)
        {
            selectionList l;
            l = lb->list;
            newtListboxAddEntry(listbox, l->title, l->data);
            lb->list = l->next;
            free(l);
        }
	newtListboxSetCurrent(listbox,lb->def);
        *(lb->comp) = listbox;
        free(lb);

        newtFormAddComponent(form, listbox);
    }
    else /*   when we actually are dealing with a newtComponent.   */
    {
        /*   add to the form.   */
        newtComponent comp = (newtComponent)group;

        comp->left = x;
        comp->top = y;
        newtFormAddComponent(form, comp);
    }
}


newtComponent askUser(enum open_method method, char * title, 
    newtComponent focus, componentGroup group)
{
    newtComponent form;
    int title_len, win_width;

    /*   open the window and make up a dialog.   */
    title_len = strlen(title);
    if(abs(group->width) < title_len)
    {
        win_width = title_len;
        if(group->width < 0)
        {
            group->width = -title_len;
        }
    }
    else
    {
        win_width = abs(group->width);
    }
    openCenteredWindow(win_width + 2, abs(group->height) + 1, title);

    /*   add all the components into a form.   */
    formAddGroup(form = newtForm(NULL, NULL, 0), group, 1, 1);

    /*   depending on what the caller wanted, we just show the
         form, or wait for the user to fill the form.   */
    if(method == show)
    {
        /*   we'll just show the form and return.   */
        newtDrawForm(form);
        newtRefresh();

        /*   clean up.   */
        newtFormDestroy(form);

        return NULL;
    }
    else if(method == layout)
    {
        // we did what we had to do.  just return.
        return form;
    }
    else if(method == ask)
    {
        /*   we'll ask the user to fill the form.   */
        newtComponent answer;
        if(focus)
        {
            newtFormSetCurrent(form, focus);
        }
        answer = newtRunForm(form);
        if(answer == form)
        {
            answer = newtFormGetCurrent(form);
        }

        newtFormDestroy(form);
        newtPopWindow();

        return answer;
    }
    else // method should be ask_keep.
    {
        // we'll ask the user, but don't try to interpret the result.
        if(focus)
        {
            newtFormSetCurrent(form, focus);
        }
        newtRunForm(form);

        return form;
    }
}


/*
//////////////////////////////////////////////////////////////////////
// specific window functions
//////////////////////////////////////////////////////////////////////
*/

/*
// this one is supposed to be used by other modules to display
// messages.
*/
void msgWindow(char *title, char *buttontext, char *message, ...)
{
    char *buf = NULL;
    int size = 10000;
    va_list args;

    /*
    // create the text first.
    // this could be a redundant work since we'll call properText
    // later.  because of the difficulty of args handling, i had
    // to do this.
    */
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    askUser(ask, title, NULL,
        compGroup(vertical, center,
            properTextbox(buf),
            groupMargine(0, 1, compGroup(horizontal, bottomRight,
                newtButton(0, 0, buttontext ? buttontext : _("Ok")))),
            NULL));
    free(buf);
}

/*
// this one show the message and return immediately without waiting
// for the user to input anything.
*/
void statusWindow(char *title, char *message, ...)
{
    char *buf = NULL;
    int size = 10000;
    va_list args;

    /*
    // create the text first.
    // this could be a redundant work since we'll call properText
    // later.  because of the difficulty of args handling, i had
    // to do this.
    */
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    askUser(show, title, NULL, (componentGroup)properTextbox(buf));
    free(buf);
}

/*
// this one is for error messages.
*/
void errWindow(bool addErrorStr, char *message, ...)
{
    char *buf = NULL;
    int size = 10000;
    va_list args;

/*
    // create the text first.
    // this could be a redundant work since we'll call properText
    // later.  because of the difficulty of args handling, i had
    // to do this.
		     */
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    // add the system error string.
    if(addErrorStr)
    {
        // catenate formatting string to the just created buffer.
        buf = realloc(buf, strlen(buf) + 5);
        strcat(buf, ": %s");
    }

    // open the dialog.
    askUser(ask, _("Error"), NULL, 
        compGroup(vertical, center,
            addErrorStr ? 
                properTextbox(buf, strerror(errno)):
                properTextbox(buf),
            groupMargine(0, 1, compGroup(horizontal, bottomRight,
                newtButton(0, 0, _("Ok")))), 
            NULL));
    free(buf);
}


/* // asks a yes/no question. */
bool yesnoWindow(char *title, bool expectYes, 
    char *btext1, char *btext2, char *message, ...)
{
    bool ret;
    newtComponent yes, no;
    char *buf = NULL;
    int size = 10000;
    va_list args;

    /*
    // create the text first.
    // this could be a redundant work since we'll call properText
    // later.  because of the difficulty of args handling, i had
    // to do this.
    */
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    yes = newtButton(0, 0, btext1 ? btext1 : _("Yes"));
    no  = newtButton(0, 0, btext2 ? btext2 : _("No"));
    ret = (askUser(ask, title, expectYes ? yes : no,
        compGroup(vertical, center, 
            properTextbox(buf),
            groupMargine(0, 1, compGroup(horizontal, bottomRight, 
                yes,
                no,
                NULL)),
            NULL)) == yes);

    free(buf);
    return ret;
}

/*
// get a single line of text which as to be free'ed.
*/
char *entryWindow(char *title, int entry_len, char *def_text,
    char *btext1, char *btext2, char *message, ...)
{
    newtComponent form, ok, cancel, entry;
    char *entry_buf;
    char *buf = NULL;
    int size = 10000;
    va_list args;
/*
    // create the text first.
    // this could be a redundant work since we'll call properText
    // later.  because of the difficulty of args handling, i had
    // to do this.
    */
    va_start(args, message);
    buf = malloc(size);
    while(vsnprintf(buf, size, message, args) == size)
    {
        size += 1000;
        free(buf);
        buf = malloc(size);
    }
    va_end(args);

    form = askUser(ask_keep, title, entry, 
        compGroup(vertical, center,
            groupMargine(0, 1, compGroup(horizontal, topLeft,
                properTextbox(buf),
                NULL)),
            entry = newtEntry(0, 0, def_text ? def_text : "", 
                    entry_len, &entry_buf, 0),
            groupMargine(0, 1, compGroup(horizontal, bottomRight,
                ok     = newtButton(0, 0, btext1 ? btext1 : _("Ok")),
                cancel = newtButton(0, 0, btext2 ? btext2 : _("Cancel")),
                NULL)),
            NULL));

    free(buf);
    if(newtFormGetCurrent(form) == cancel)
    {
        releaseForm(&form);
        return NULL;
    }

    entry_buf = strdup(entry_buf);
    releaseForm(&form);
    return entry_buf;
}



